﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.OleDb;
using System.Data;
using System.Xml.Linq;
using System.IO;
using HP.FOT.Shared.LoggingSupport;
using System.Reflection;
using Eds.AirSoa.Util.Messaging;
using Microsoft.CSharp;
using System.CodeDom.Compiler;
using System.CodeDom;

namespace RefactorMessages
{
    /// <remarks>
    /// To convert Message resource XML to spreadsheet for analysis
    /// </remarks>
    public class AnalyzeMessageXmlHelper
    {
        List<InternalMessageKeyEntity> _allXMLKeys = new List<InternalMessageKeyEntity>();
        string _tower;
        string _fileMessageKeysCsvPath;
        string _fileMessageClassCsvPath;
        string _xmlPath;
        string _classPath;

        public AnalyzeMessageXmlHelper(string outputFolder, string srcFolder, string tower, string language = "English")
        {
            this._tower = tower;

            this._fileMessageKeysCsvPath = outputFolder + "\\" + tower.ToString() + "MessageKeys.csv";
            this._fileMessageClassCsvPath = outputFolder + "\\" + tower.ToString() + "MessageClass.csv";

            this._xmlPath = srcFolder + "\\" + language + "\\" + tower.ToString() + "Messages.xml";
            this._classPath = srcFolder + "\\" + tower.ToString() + "MessageKeys.cs";
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="excelFilePath">Spreadsheet file path</param>
        /// <param name="outputXmlFilePath">Output XML file path</param>
        public void GenerateXmlFromExcel(string excelFilePath, string sheetName = "Finalized")
        {
            var msg = PrepareObjectFromExcel(excelFilePath, sheetName);

            // write Messages.xml
            string xmlContext = Util.Serialize2XML(msg);
            Util.WriteFile(_xmlPath, xmlContext);

            // write Class
            this.GenerateClass(msg);
        }

        public void GenerateCsvFromMessageResourceXml(bool includeDetect = true)
        {
            if (this._allXMLKeys == null || this._allXMLKeys.Count <= 0)
            {
                this.PrepareMessageXMLKeys();
            }

            var dt = GenerateDataTableFromXml(includeDetect);
            List<string> lstItem = new List<string>();
            StringBuilder sbCSVContent = new StringBuilder();

            lstItem.Clear();
            string strCSVHeader = string.Empty;
            foreach (DataColumn dc in dt.Columns)
            {
                lstItem.Add(dc.ColumnName);
            }
            strCSVHeader = string.Join(",", lstItem);
            sbCSVContent.AppendLine(strCSVHeader);

            foreach (DataRow dr in dt.Rows)
            {
                lstItem.Clear();
                foreach (DataColumn dc in dt.Columns)
                {
                    lstItem.Add(string.Format("\"{0}\"", dr[dc.ColumnName].ToString().Replace(@"""", @"""""")));
                }
                sbCSVContent.AppendLine(string.Join(",", lstItem.ToArray()));
            }

            if (File.Exists(_fileMessageKeysCsvPath))
            {
                File.Delete(_fileMessageKeysCsvPath);
            }
            File.WriteAllText(_fileMessageKeysCsvPath, sbCSVContent.ToString());
        }

        public void GenerateClass(Messages src)
        {
            GenerateCodeMessageKeyClass generator = new GenerateCodeMessageKeyClass(src, _tower);
            generator.Generate(_classPath);
        }

        #region Private Methods
        /// <summary>
        /// Prepare DataTable for exporting to CSV
        /// </summary>
        /// <param name="includeDetect"></param>
        /// <returns></returns>
        private DataTable GenerateDataTableFromXml(bool includeDetect = true)
        {
            DataTable dt = null;

            Encoding ec = new UTF8Encoding(true);
            using (StreamReader sr = new StreamReader(_xmlPath, ec))
            {
                XDocument document = XDocument.Load(sr);

                var qryTypes = from type in document.Descendants(@"MessagesByType")
                               select new
                               {
                                   TypeCode = type.Attribute("typeCode"),
                                   TypeName = type.Attribute("typeName")
                               };

                var query = from msg in document.Descendants(@"MessagesByType").Descendants(@"Message")
                            select new
                            {
                                TypeCode = msg.Parent.Attribute("typeCode").Value,
                                TypeName = msg.Parent.Attribute("typeName").Value,
                                MessageId = msg.Attribute("id").Value,
                                //OldMessageId = _tower == TowerEnum.AM ? msg.Attribute("old_id").Value : string.Empty,
                                //Category = _tower == TowerEnum.AM ? msg.Attribute("category").Value : string.Empty,
                                EventID = msg.Element("EventID") == null ? string.Empty : msg.Element("EventID").Value,
                                Code = msg.Element("Code") == null ? string.Empty : msg.Element("Code").Value,
                                ShortText = msg.Element("ShortText") == null ? string.Empty : msg.Element("ShortText").Value,
                                Description = msg.Element("Description") == null ? string.Empty : msg.Element("Description").Value,
                                LongST = msg.Element("ShortText") != null && msg.Element("ShortText").Value.Length > 64 ? (msg.Element("ShortText").Value.Length - 64).ToString() : string.Empty,
                                STIncludeParam = msg.Element("ShortText") != null && msg.Element("ShortText").Value.IndexOf("{") > 0 ? "X" : string.Empty,
                                BeVAL = msg.Attribute("id") != null && msg.Attribute("id").Value.IndexOf("V") > 0 ? "X" : string.Empty,
                                MissingVAR = msg.Element("Description").Value != null && msg.Element("Description").Value.IndexOf("{") > 0 && msg.Element("Variables") == null ? "X" : string.Empty,
                                Occurs = _allXMLKeys.Count(o => o.MessageId == msg.Attribute("id").Value),
                                Variables = GetMessageVariables(msg.Element("Variables"), msg.Elements("Variable").ToList()),
                                KeyNameInClass = string.Join("\r\n", _allXMLKeys.Where(o => o.MessageId == msg.Attribute("id").Value).Select(o => o.ClassName).ToList())
                            };

                var list = query.ToList();

                dt = new DataTable();
                dt.Columns.Add("Type");
                if (string.Compare(_tower, TowerEnum.AM.ToString(), true) == 0)
                {
                    //dt.Columns.Add("Category");
                    //dt.Columns.Add("OldMessageId");
                }
                dt.Columns.Add("MessageId");
                dt.Columns.Add("EventID");
                dt.Columns.Add("Code");
                dt.Columns.Add("ShortText");
                dt.Columns.Add("Description");
                dt.Columns.Add("Variables");
                if (includeDetect)
                {
                    dt.Columns.Add("LongST");
                    dt.Columns.Add("STIncludeParam");
                    dt.Columns.Add("BeVAL");
                    dt.Columns.Add("MissingVAR");
                    dt.Columns.Add("Occurs");
                }
                dt.Columns.Add("KeyNameInClass");

                foreach (var o in list)
                {
                    DataRow dr = dt.NewRow();
                    dr["Type"] = o.TypeName;
                    if (string.Compare(_tower, TowerEnum.AM.ToString(), true) == 0)
                    {
                        //dr["Category"] = o.Category;
                        //dr["OldMessageId"] = o.OldMessageId;
                    }
                    dr["MessageId"] = o.MessageId;
                    dr["EventID"] = o.EventID;
                    dr["Code"] = o.Code;
                    dr["ShortText"] = o.ShortText;
                    dr["Description"] = o.Description;
                    dr["Variables"] = o.Variables;
                    if (includeDetect)
                    {
                        dr["LongST"] = o.LongST;
                        dr["STIncludeParam"] = o.STIncludeParam;
                        dr["BeVAL"] = o.BeVAL;
                        dr["MissingVAR"] = o.MissingVAR;
                        dr["Occurs"] = o.Occurs;
                    }
                    dr["KeyNameInClass"] = o.KeyNameInClass;

                    dt.Rows.Add(dr);
                }
            }

            return dt;
        }

        /// <summary>
        /// Get Message (XML)'s variables
        /// </summary>
        /// <param name="vars"></param>
        /// <param name="varitems"></param>
        /// <returns></returns>
        private string GetMessageVariables(XElement vars, List<XElement> varitems)
        {
            string ret = string.Empty;
            List<XElement> items = null;

            if (varitems != null) items = varitems;
            if (vars != null) items = vars.Elements("Variable").ToList();

            if (items != null && items.Count > 0)
            {
                ret = string.Join("\r\n", items.Select(p => p.Attribute("formatIndex").Value + " " + p.Value).ToList());
            }
            return ret;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="mid"></param>
        /// <returns></returns>
        private static string GetCodeFromMessageId(string mid)
        {
            if (!string.IsNullOrEmpty(mid) && mid.IndexOf("_") > 0)
            {
                return mid.Split('_')[1] + "." + mid.Split('_')[0];
            }
            else
            {
                return "[ERROR!]";
            }
        }

        /// <summary>
        /// Read Message's Variables from XML
        /// </summary>
        /// <param name="content"></param>
        /// <returns></returns>
        private static List<Variable> ReadVariables(string content)
        {
            if (string.IsNullOrEmpty(content)) return null;

            string d = "\n";
            var query = from item in content.Split(d.ToCharArray())
                        select
                            new Variable
                            {
                                formatIndex = item.Substring(0, item.IndexOf(' ')),
                                text = item.Substring(item.IndexOf(' ') + 1, item.Length - item.IndexOf(' ') - 1)
                            };

            return query.ToList();
        }

        /// <summary>
        /// Shall be more automatical/configurable
        /// </summary>
        private void PrepareMessageXMLKeys()
        {
            Type VAL = typeof(AMMessageKeys.ValidationErrorMessages);
            Type APP = typeof(AMMessageKeys.ApplicationErrorMessages);
            Type SYS = typeof(AMMessageKeys.SystemErrorMessages);
            Type WRN = typeof(AMMessageKeys.WarnMessages);
            Type INF = typeof(AMMessageKeys.InfoMessages);

            StringBuilder sb = new StringBuilder();
            if (VAL != null) AggregateMessageKeyClass(ref sb, VAL, "Validation Error");
            if (APP != null) AggregateMessageKeyClass(ref sb, APP, "Application Error");
            if (SYS != null) AggregateMessageKeyClass(ref sb, SYS, "System Error");
            if (WRN != null) AggregateMessageKeyClass(ref sb, WRN, "Warn");
            if (INF != null) AggregateMessageKeyClass(ref sb, INF, "Info");

#if DEBUG_Retired
            string ret = sb.ToString();
            if (File.Exists(_fileMessageClassCsvPath))
            {
                File.Delete(_fileMessageClassCsvPath);
            }
            File.WriteAllText(_fileMessageClassCsvPath, sb.ToString());
#endif
        }

        /// <summary>
        /// To aggregate all message keys of all message types since they're different classes
        /// </summary>
        /// <param name="sbAggregation"></param>
        /// <param name="messageType"></param>
        /// <param name="typeName"></param>
        private void AggregateMessageKeyClass(ref StringBuilder sbAggregation, Type messageType, string typeName)
        {
            // 1. retrieve message keys, generate file
            var variables = messageType.GetFields(BindingFlags.Instance | BindingFlags.Static | BindingFlags.Public)
                .ToDictionary(f => f.Name, f => f.GetValue(f));

            foreach (var o in variables)
            {
                MessageKey ov = o.Value as MessageKey;
                if (ov != null)
                {
                    InternalMessageKeyEntity mk = new InternalMessageKeyEntity
                    {
                        TypeName = ov.Type.ToString(),
                        ClassName = o.Key,
                        MessageId = ov.Code
                    };
                    _allXMLKeys.Add(mk);
                    sbAggregation.AppendLine(mk.TypeName + "," + mk.ClassName + "," + mk.MessageId);
                }
            }
        }

        private Messages PrepareObjectFromExcel(string excelFilePath, string sheetName)
        {
            var fileName = excelFilePath;
            var connectionString = string.Format("Provider=Microsoft.Jet.OLEDB.4.0; data source={0}; Extended Properties=Excel 8.0;", fileName);

            var adapter = new OleDbDataAdapter("SELECT * FROM [" + sheetName + "$]", connectionString);
            var ds = new DataSet();

            adapter.Fill(ds, sheetName);
            var data = ds.Tables[sheetName].AsEnumerable();

            var query = data.Select(x =>
                new Message
                {
                    TypeName = x.Field<string>("Type Name"),
                    TypeCode = x.Field<string>("Type Code"),
                    ClassName = x.Field<string>("Key Name In Class"),
                    old_id = x.Field<string>("OldMessageId"),
                    category = x.Field<string>("Category"),
                    id = x.Field<string>("MessageId"),
                    EventId = x.Field<string>("EventID"),
                    Code = GetCodeFromMessageId(x.Field<string>("MessageId")),
                    ShortText = x.Field<string>("ShortText"),
                    Description = x.Field<string>("Description"),
                    Variables = ReadVariables(x.Field<string>("Variables")),
                }).ToList<Message>().GroupBy(o => new { o.TypeCode, o.TypeName });

            Messages msg = new Messages();
            var msgsByTypes = new List<MessagesByType>();

            foreach (var type in query)
            {
                MessagesByType msgBody = new MessagesByType
                {
                    typeCode = type.Key.TypeCode,
                    typeName = type.Key.TypeName,
                    Messages = type.ToList()
                };

                msgsByTypes.Add(msgBody);
            }

            msg.MessagesByTypes = msgsByTypes;

            return msg;
        }
        #endregion Private
    }
}
