﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO.Compression;
using System.Xml.Serialization;

namespace RefactorMessages
{
    /// <summary>
    /// from TFS\FOTFP\R2\Common\CBP\HP.FOT.CBP.Shared\Helpers\Serializer.cs
    /// </summary>
    public class Serializer
    {
        const int STRING_MAX_LENGTH = 16 * 1024 + 256;//the length of any string object < 16K

        #region for binaryformatter serialize
        public byte[] SerializeObject(object serializableObject)
        {
            MemoryStream stream = new MemoryStream();
            BinaryFormatter b = new BinaryFormatter();
            b.Serialize(stream, serializableObject);
            byte[] bytes = stream.GetBuffer();
            stream.Close();
            return bytes;

        }

        /// <summary>
        /// 
        /// </summary>
        /// <typeparam name="T">the type of the object need to be serialized</typeparam>
        /// <param name="serializableObject"></param>
        /// <returns></returns>
        public static string SerializeObjectToString<T>(T serializableObject)
        {
            Serializer serializer = new Serializer();
            byte[] bytes = serializer.SerializeObject(serializableObject);

            // SEP this was causing problems with deserialization so I commented it out
            //bytes = Compress(bytes);

            return ToString(bytes);
        }

        #endregion

        #region for binaryformatter deserialize
        public object DeSerializeObject(byte[] someBytes)
        {
            MemoryStream stream = new MemoryStream(someBytes);
            BinaryFormatter b = new BinaryFormatter();
            object retObj = b.Deserialize(stream);
            stream.Close();
            return retObj;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <typeparam name="T">the type of the object got by deserialized</typeparam>
        /// <param name="serializedObjectBytesStr"></param>
        /// <returns></returns>
        public static T DeSerializeObject<T>(string serializedObjectBytesStr)
        {
            T retObj;

            try
            {
                byte[] bytes = ToBytes(serializedObjectBytesStr);

                // SEP this was causing problems with deserialization so I commented it out
                //bytes = Decompress(bytes);

                Serializer serializer = new Serializer();
                retObj = (T)serializer.DeSerializeObject(bytes);
            }
            catch
            {
                retObj = default(T);
            }

            return retObj;
        }

        #endregion

        public static string SerializeObject2XML(object serializableObject)
        {
            if (serializableObject == null) return null;

            return SerializeObject2XML(serializableObject, serializableObject.GetType());
        }

        public static string SerializeObject2XML(object serializableObject, Type type)
        {
            if (serializableObject == null) return null;

            if (type == null) type = serializableObject.GetType();

            System.Xml.XmlWriterSettings settings = new System.Xml.XmlWriterSettings();
            settings.OmitXmlDeclaration = true;
            settings.Indent = true;

            System.Text.StringBuilder builder = new System.Text.StringBuilder();
            System.Xml.XmlWriter xw = System.Xml.XmlWriter.Create(builder, settings);

            System.Xml.Serialization.XmlSerializer x = new System.Xml.Serialization.XmlSerializer(serializableObject.GetType());
            x.Serialize(xw, serializableObject);

            return builder.ToString();
        }

        /// <summary>
        /// Deserializes workflow markup into an T object
        /// </summary>
        /// <param name="xml">string workflow markup to deserialize</param>
        /// <param name="obj">Output T object</param>
        /// <param name="exception">output Exception value if deserialize failed</param>
        /// <returns>true if this XmlSerializer can deserialize the object; otherwise, false</returns>
        public static bool Deserialize<T>(string xml, out T obj, out System.Exception exception)
        {
            exception = null;
            obj = default(T);
            try
            {
                obj = Deserialize<T>(xml);
                return true;
            }
            catch (System.Exception ex)
            {
                exception = ex;
                return false;
            }
        }

        public static bool Deserialize<T>(string xml, out T obj)
        {
            System.Exception exception = null;
            return Deserialize(xml, out obj, out exception);
        }

        public static T Deserialize<T>(string xml)
        {
            System.IO.StringReader stringReader = null;
            try
            {
                XmlSerializer serializer = new XmlSerializer(typeof(T));
                stringReader = new System.IO.StringReader(xml);

                return ((T)(serializer.Deserialize(System.Xml.XmlReader.Create(stringReader))));
            }
            finally
            {
                if ((stringReader != null))
                {
                    stringReader.Dispose();
                }
            }
        }

        #region private method
        private static string ToString(byte[] ms)
        {
            return Convert.ToBase64String(ms);
        }

        private static byte[] ToBytes(string serializedObj)
        {
            return Convert.FromBase64String(serializedObj);
        }

        private static byte[] Compress(byte[] buffer)
        {
            MemoryStream ms = new MemoryStream();
            DeflateStream stream = new DeflateStream(ms, CompressionMode.Compress, true);
            stream.Write(buffer, 0, buffer.Length);
            stream.Close();

            buffer = ms.ToArray();
            ms.Close();

            return buffer;
        }

        private static byte[] Decompress(byte[] buffer)
        {
            MemoryStream ms = new MemoryStream();
            ms.Write(buffer, 0, buffer.Length);
            ms.Position = 0;
            DeflateStream stream = new DeflateStream(ms, CompressionMode.Decompress);
            stream.Flush();

            byte[] decompressBuffer = new byte[STRING_MAX_LENGTH];
            int nSizeIncept = stream.Read(decompressBuffer, 0, STRING_MAX_LENGTH);
            stream.Close();
            ms.Close();

            byte[] lastResult = new byte[nSizeIncept];
            System.Buffer.BlockCopy(decompressBuffer, 0, lastResult, 0, nSizeIncept);

            return lastResult;
        }
        #endregion
    }
}

